'use strict';

var _jestEach = require('jest-each');

var _jestUtil = require('jest-util');

var _state = require('./state');

/**
 * Copyright (c) 2014-present, Facebook, Inc. All rights reserved.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 *  strict-local
 */

const describe = (blockName, blockFn) =>
  _dispatchDescribe(blockFn, blockName, describe);
describe.only = (blockName, blockFn) =>
  _dispatchDescribe(blockFn, blockName, describe.only, 'only');
describe.skip = (blockName, blockFn) =>
  _dispatchDescribe(blockFn, blockName, describe.skip, 'skip');

const _dispatchDescribe = (blockFn, blockName, describeFn, mode) => {
  const asyncError = new _jestUtil.ErrorWithStack(undefined, describeFn);
  if (blockFn === undefined) {
    asyncError.message = `Missing second argument. It must be a callback function.`;
    throw asyncError;
  }
  if (typeof blockFn !== 'function') {
    asyncError.message = `Invalid second argument, ${blockFn}. It must be a callback function.`;
    throw asyncError;
  }
  (0, _state.dispatch)({
    asyncError: asyncError,
    blockName: blockName,
    mode: mode,
    name: 'start_describe_definition'
  });
  blockFn();
  (0, _state.dispatch)({
    blockName: blockName,
    mode: mode,
    name: 'finish_describe_definition'
  });
};

const _addHook = (fn, hookType, hookFn, timeout) => {
  const asyncError = new _jestUtil.ErrorWithStack(undefined, hookFn);

  if (typeof fn !== 'function') {
    asyncError.message =
      'Invalid first argument. It must be a callback function.';

    throw asyncError;
  }

  (0, _state.dispatch)({
    asyncError: asyncError,
    fn: fn,
    hookType: hookType,
    name: 'add_hook',
    timeout: timeout
  });
};

// Hooks have to pass themselves to the HOF in order for us to trim stack traces.
const beforeEach = (fn, timeout) =>
  _addHook(fn, 'beforeEach', beforeEach, timeout);
const beforeAll = (fn, timeout) =>
  _addHook(fn, 'beforeAll', beforeAll, timeout);
const afterEach = (fn, timeout) =>
  _addHook(fn, 'afterEach', afterEach, timeout);
const afterAll = (fn, timeout) => _addHook(fn, 'afterAll', afterAll, timeout);

const test = (testName, fn, timeout) =>
  _addTest(testName, undefined, fn, test, timeout);
const it = test;
test.skip = (testName, fn, timeout) =>
  _addTest(testName, 'skip', fn, test.skip, timeout);
test.only = (testName, fn, timeout) =>
  _addTest(testName, 'only', fn, test.only, timeout);
test.todo = function(testName) {
  if (
    (arguments.length <= 1 ? 0 : arguments.length - 1) > 0 ||
    typeof testName !== 'string'
  ) {
    throw new _jestUtil.ErrorWithStack(
      'Todo must be called with only a description.',
      test.todo
    );
  }
  return _addTest(testName, 'todo', () => {}, test.todo);
};

const _addTest = (testName, mode, fn, testFn, timeout) => {
  const asyncError = new _jestUtil.ErrorWithStack(undefined, testFn);

  if (typeof testName !== 'string') {
    asyncError.message = `Invalid first argument, ${testName}. It must be a string.`;

    throw asyncError;
  }
  if (fn === undefined) {
    asyncError.message =
      'Missing second argument. It must be a callback function. Perhaps you want to use `test.todo` for a test placeholder.';

    throw asyncError;
  }
  if (typeof fn !== 'function') {
    asyncError.message = `Invalid second argument, ${fn}. It must be a callback function.`;

    throw asyncError;
  }

  return (0, _state.dispatch)({
    asyncError: asyncError,
    fn: fn,
    mode: mode,
    name: 'add_test',
    testName: testName,
    timeout: timeout
  });
};

test.each = (0, _jestEach.bind)(test);
test.only.each = (0, _jestEach.bind)(test.only);
test.skip.each = (0, _jestEach.bind)(test.skip);

describe.each = (0, _jestEach.bind)(describe, false);
describe.only.each = (0, _jestEach.bind)(describe.only, false);
describe.skip.each = (0, _jestEach.bind)(describe.skip, false);

module.exports = {
  afterAll: afterAll,
  afterEach: afterEach,
  beforeAll: beforeAll,
  beforeEach: beforeEach,
  describe: describe,
  it: it,
  test: test
};
